#include "fw/dynamixel/context.h"
#include "shared/protocol.h"

#pragma once

namespace schpin
{

using dynmx_servo_events_variant = std::variant< float >;

class dynmx_servo
{
        struct idle_task
        {
        };

        struct activate_task
        {
                std::size_t i;
        };

        using task_variant = std::variant< idle_task, activate_task >;

        dynmx_context context_;
        task_variant  task_ = { activate_task{} };

        public:
        using receive_type   = dynmx_master_tuple;
        using events_variant = dynmx_servo_events_variant;

        explicit dynmx_servo( dynmx_servo_id id )
          : context_( id )
        {
        }

        servo_id get_id() const
        {
                return *context_.get_id();
        }

        bool has_task() const
        {
                return !std::holds_alternative< idle_task >( task_ );
        }

        bool is_activating() const
        {
                return std::holds_alternative< activate_task >( task_ );
        }

        template < typename Visitor >
        void stop( Visitor&& )
        {
        }

        std::optional< servo_master_message > on_servo_receive( clk_time,
                                                                s_ctl_id            sid,
                                                                const dynmx_buffer& buff )
        {

                context_.on_receive( sid, buff );

                return {};
        }

        template < typename Visitor >
        void tick( clk_time, Visitor&& vis )
        {
                task_ << match( [&]( idle_task ) {},
                                [&]( activate_task at ) {
                                        at.i = context_.full_poll( at.i, vis );
                                        if ( at.i == 0 ) {
                                                task_ = idle_task{};
                                        } else {
                                                task_ = at;
                                        }
                                } );
        }

        template < typename Visitor >
        void on_event( ctl_id, clk_time, dynmx_servo_events_variant, Visitor&& )
        {
        }

        template < typename Visitor >
        void poll( clk_time, std::size_t, Visitor&& )
        {
        }

        em::either< dynmx_servo_events_variant, servo_master_message >
            on_servo_packet( ctl_id, master_servo_command_variant )
        {
                return { dynmx_servo_events_variant{ 0.f } };
        }

        void dump_info()
        {
                context_.dump_info();
        }
};
}  // namespace schpin
