#include "fw/lewan/servo.h"
#include "test/servo.h"

#pragma once

namespace schpin
{

class lewan_servo_tester
{
        lewan_servo servo_;
        sim_servo   sim_;

        template < typename Visitor >
        auto make_matcher( clk_time ctime, Visitor&& vis )
        {
                return match(
                    [&]( servo_tag, s_ctl_id msg_id, reply, const master_lewan_message& m ) {
                            master_lewan_protocol::extract( m )
                                .bind_left(
                                    em::uncurry |
                                    [&]( lewan_servo_id id, master_lewan_submessage submsg ) {
                                            return master_lewan_group::extract( submsg )
                                                .convert_left( [&]( auto var ) {
                                                        return std::make_tuple( id, var );
                                                } );
                                    } )
                                .match(
                                    em::uncurry |
                                        [&]( lewan_servo_id id, master_lewan_variant var ) {
                                                pass_to_sim( ctime, msg_id, id, var, vis );
                                        },
                                    [&]( auto ) {} );
                    },
                    [&]( control_tag, servo_id id, servo_master_message m ) {
                            vis( control_tag{}, id, m );
                    },
                    [&]( activate_tag, servo_id id ) {
                            vis( activate_tag{}, id );
                    } );
        }

        template < typename Visitor >
        void pass_to_sim( clk_time                    ctime,
                          s_ctl_id                    s_id,
                          lewan_servo_id              id,
                          const master_lewan_variant& var,
                          Visitor&&                   vis )
        {
                auto opt_msg = sim_.on_cmd( ctime, id, var );
                if ( !opt_msg ) {
                        return;
                }
                // TODO: this duplicates servogroup
                // maybe make ::extract uplift the message?
                lewan_master_protocol::extract( *opt_msg )
                    .bind_left( em::uncurry |
                                [&]( lewan_servo_id, lewan_master_submessage m ) {
                                        return lewan_master_group::extract( m ).convert_left(
                                            [&]( auto var ) {
                                                    return std::make_tuple( id, var );
                                            } );
                                } )
                    .match(
                        em::uncurry |
                            [&]( lewan_servo_id, lewan_master_variant msg_var ) {
                                    auto opt_msg = servo_.on_servo_receive( ctime, s_id, msg_var );
                                    if ( !opt_msg ) {
                                            return;
                                    }
                                    vis( control_tag{}, *id, *opt_msg );
                            },
                        [&]( auto ) {} );
        }

      public:
        using receive_type   = lewan_master_group;
        using events_variant = lewan_servo_events_variant;

        lewan_servo_tester( lewan_servo_id id, clk_time now )
          : servo_( id, now )
        {
        }

        servo_id get_id() const
        {
                return servo_.get_id();
        }

        bool has_task() const
        {
                return servo_.has_task();
        }

        bool is_activating() const
        {
                return servo_.is_activating();
        }

        template < typename Visitor >
        void stop( Visitor&& v )
        {
                // TODO: clk_time is problem
                servo_.stop( make_matcher( clk_time{}, std::forward< Visitor >( v ) ) );
        }

        std::optional< servo_master_message > on_servo_receive( clk_time,
                                                                s_ctl_id,
                                                                const lewan_master_group& )
        {
                // TODO: when this is called it is error
                return {};
        }

        template < typename Visitor >
        void tick( clk_time ctime, Visitor&& vis )
        {
                servo_.tick( ctime, make_matcher( ctime, std::forward< Visitor >( vis ) ) );
        }

        template < typename Visitor >
        void
        on_event( ctl_id cid, clk_time ctime, const lewan_servo_events_variant& var, Visitor&& vis )
        {
                servo_.on_event(
                    cid, ctime, var, make_matcher( ctime, std::forward< Visitor >( vis ) ) );
        }

        template < typename Visitor >
        void poll( clk_time ctime, std::size_t offset_i, Visitor&& vis )
        {
                servo_.poll(
                    ctime, offset_i, make_matcher( ctime, std::forward< Visitor >( vis ) ) );
        }

        em::either< lewan_servo_events_variant, servo_master_message >
        on_servo_packet( ctl_id cid, const master_servo_command_variant& var )
        {
                return servo_.on_servo_packet( cid, var );
        }

        void dump_info()
        {
                servo_.dump_info();
        }
};
}  // namespace schpin
