
#include "node/components/context.h"

using namespace schpin;

context_component::context_component( ros::NodeHandle             handle,
                                      robot_model                 rob,
                                      const node_hardware_config& c )
  : state_pub_( handle.advertise< sensor_msgs::JointState >( c.joint_state_topic, 100 ) )
  , context_( node_context{ std::move( rob ), servo_group{ c.servos } } )
{
        b_config_->input_buffer_telem_period_ms = clk_time{ 2000 };
        b_config_->event_buffer_telem_period_ms = clk_time{ 2000 };
}

void context_component::tick( const node_context&, sm_feedback_f msg_cb )
{
        if ( !b_config_.has_loaded() ) {
                b_config_.load_config( [&]( node_fw_config_message m ) {
                        msg_cb( node_fw_group::create< CONFIG >( m ) );
                } );
        }

        SCHPIN_INFO_LOG_THROTTLE( 20, "node", "Body pose: " << context_.rstate.body_pose );

        geometry_msgs::TransformStamped msg;

        msg.header.stamp          = ros::Time::now();
        msg.header.frame_id       = "world";
        msg.child_frame_id        = "base_link";
        msg.transform.translation = vector_to_msg( context_.rstate.body_pose.position );
        msg.transform.rotation    = quaternion_to_msg( context_.rstate.body_pose.orientation );

        br_.sendTransform( msg );
}

void context_component::on_feedback( fw_node_group_variant v,
                                     const node_context&,
                                     sm_feedback_f msg_cb )
{
        v << tuple_match(
            [&]( cmd_tag< CONFIG >, fw_node_config_variant var ) {
                    b_config_.on_servo_config_message( var, [&]( node_fw_config_message m ) {
                            msg_cb( node_fw_group::create< CONFIG >( m ) );
                    } );
            },
            [&]( cmd_tag< SERVO >, servo_id id, servo_master_command_variant var ) {
                    on_servo_feedback_cmd( id, var, msg_cb );
            },
            [&]( auto... ) {} );
}

