#include "node/context.h"
#include "node/state_machines/base.h"

#include <geometry_msgs/TransformStamped.h>
#include <ros/ros.h>
#include <sensor_msgs/JointState.h>
#include <tf2_ros/transform_broadcaster.h>

#pragma once

namespace schpin
{

class context_component
{
        ros::Publisher                  state_pub_;
        node_context                    context_;
        prot_config_loader< fw_config > b_config_;
        tf2_ros::TransformBroadcaster   br_;

      public:
        context_component( ros::NodeHandle handle, robot_model rob, const node_hardware_config& c );

        const node_context& get_context()
        {
                return context_;
        }

        void record_msg_sent( ctl_id cid, node_fw_message m )
        {
                context_.hist.insert( cid, m );
        }

        void feedback_processed( ctl_id cid )
        {
                context_.hist.drop( cid );
        }

        void tick( const node_context&, sm_feedback_f msg_cb );

        void set_body_pose( const pose< world_frame >& p )
        {
                context_.rstate.body_pose = p;
        }

        void set_servo_config( servo_id id, const nlohmann::json& j, sm_feedback_f vis )
        {
                context_.sgroup.set_servo_config( id, j, vis );
        }

        void on_feedback( fw_node_group_variant v, const node_context&, sm_feedback_f msg_cb );

      private:
        template < typename UnaryFunction >
        void on_servo_feedback_cmd( servo_id                     id,
                                    servo_master_command_variant v,
                                    UnaryFunction                msg_cb )
        {
                context_.sgroup.on_feedback(  //
                    id,
                    v,
                    match(
                        [&]( const servo& s, em::angle a ) {
                                bool has_updated = context_.rstate.update_joint_state(
                                    context_.robot, s.get_joint_name(), a );

                                SCHPIN_ASSERT( has_updated );

                                sensor_msgs::JointState msg;
                                msg.name.push_back( s.get_joint_name() );
                                msg.position.push_back( static_cast< double >( *a ) );

                                state_pub_.publish( msg );
                        },
                        [&]( node_fw_group_message m ) {
                                msg_cb( m );
                        } ) );
        }
};

}  // namespace schpin
