#include "fw/dynamixel/protocol.h"

#include <gtest/gtest.h>
#include <schpin_lib/view.h>

using namespace schpin;

TEST( dynamixel_protocol, master_dynmx_group )
{
        std::array< uint8_t, 4 > buffer{ 0x00, 0x02, 0x00, 0x00 };

        master_dynmx_submessage msg = master_dynmx_group::create< DYNMX_WRITE >(
            dynmx_addr{ 0x0074 }, dynmx_buffer{ buffer } );

        master_dynmx_group::extract( msg ).match(
            [&]( master_dynmx_variant var ) {
                    bool holds_right_alternative = std::holds_alternative<
                        std::tuple< cmd_tag< DYNMX_WRITE >, dynmx_addr, dynmx_buffer > >( var );
                    EXPECT_TRUE( holds_right_alternative );
            },
            [&]( protocol_error_record rec ) {
                    FAIL() << rec << "\n" << msg;
            } );
}

TEST( dynamixel_protocol, master_dynmx )
{
        std::array< uint8_t, 4 > buffer{ 0x00, 0x02, 0x00, 0x00 };

        master_dynmx_message msg = make_master_dynmx_msg< DYNMX_WRITE >(
            dynmx_servo_id::get< 1 >(), dynmx_addr{ 0x0074 }, dynmx_buffer{ buffer } );

        // TODO: fuuuuuuck clang format :/
        master_dynmx_message expected{ std::array< uint8_t, 16 >{ 0xFF,
                                                                  0xFF,
                                                                  0xFD,
                                                                  0x00,
                                                                  0x01,
                                                                  0x09,
                                                                  0x00,
                                                                  0x03,
                                                                  0x74,
                                                                  0x00,
                                                                  0x00,
                                                                  0x02,
                                                                  0x00,
                                                                  0x00,
                                                                  0xCA,
                                                                  0x89 } };

        EXPECT_EQ( msg, expected ) << std::hex << convert_view< int >( msg ) << "\n"
                                   << convert_view< int >( expected );

        master_dynmx_protocol::extract( msg ).match(
            em::uncurry | [&]( dynmx_header, dynmx_servo_id, master_dynmx_variant, dynmx_crc ) {},
            [&]( protocol_error_record rec ) {
                    FAIL() << rec << "\n" << msg;
            } );
}

TEST( dynamixel_protocol, dynmx_master )
{
        dynmx_master_message msg{ std::array< uint8_t, 12 >{
            0xff, 0xff, 0xfd, 0x0, 0x1, 0x5, 0x0, 0x55, 0x0, 0x0, 0x53, 0x21 } };

        dynmx_master_protocol::extract( msg ).match(
            em::uncurry | [&]( dynmx_header, dynmx_servo_id, auto, dynmx_crc ) {},
            [&]( auto e ) {
                    FAIL() << e << "\n" << convert_view< int >( msg );
            } );
}

int main( int argc_in, char** argv_in )
{
        testing::InitGoogleTest( &argc_in, argv_in );

        return RUN_ALL_TESTS();
}
