#include "node/lua_vm.h"

#include <schpin_lib/lua/repl.h>
#include <schpin_msgs/LuaCall.h>
#include <schpin_msgs/LuaDef.h>

using namespace schpin;

class console
{
        ros::NodeHandle    handle_;
        slua_repl          repl_;
        lua_State*         L_ptr_;
        ros::ServiceClient lua_call_;

        static int lua_callback( lua_State* l_ptr )
        {
                console* c_ptr =
                    reinterpret_cast< console* >( lua_touserdata( l_ptr, lua_upvalueindex( 1 ) ) );
                const char* name = luaL_checkstring( l_ptr, lua_upvalueindex( 2 ) );

                return c_ptr->handle_callback( name, l_ptr );
        }

        int handle_callback( const char* name, lua_State* l_ptr )
        {
                nlohmann::json args;
                for ( int i : em::range< int >( lua_gettop( l_ptr ) ) ) {
                        // TODO: make some proper handler for json in lua/def.h or json.h?
                        try {
                                args.push_back( slua_json_decode( l_ptr, i + 1 ) );
                        }
                        catch ( const nlohmann::json::exception& e ) {
                                SCHPIN_ERROR_LOG( "console", e.what() );
                        }
                }

                schpin_msgs::LuaCall srv;
                srv.request.name = name;
                srv.request.args = args.dump();

                if ( !lua_call_.call( srv ) ) {
                        SCHPIN_ERROR_LOG( "console", "Failed to call service" );
                        return 0;
                }

                if ( srv.response.success == 0u ) {
                        SCHPIN_ERROR_LOG( "console",
                                          "Remote service call failed: " << srv.response.response );
                        return 0;
                }

                nlohmann::json res = srv.response.response;
                slua_json_encode( l_ptr, res );

                // res.size() ?
                return 1;
        }

      public:
        console( const ros::NodeHandle& handle, const nlohmann::json& def )
          : handle_( handle )
          , lua_call_( handle_.serviceClient< schpin_msgs::LuaCall >( "/control/lua/call" ) )
        {

                L_ptr_ = luaL_newstate();
                luaL_openlibs( L_ptr_ );

                em::for_each( def, [&]( nlohmann::json j ) {
                        if ( j["type"] != "callback" && j["type"] != "lua procedure" ) {
                                return;
                        }
                        lua_pushlightuserdata( L_ptr_, this );
                        lua_pushstring( L_ptr_, j["name"].get< std::string >().c_str() );
                        lua_pushcclosure( L_ptr_, lua_callback, 2 );
                        lua_setglobal( L_ptr_, j["name"].get< std::string >().c_str() );
                } );
        }

        void run()
        {
                while ( true ) {

                        std::string msg = repl_.prompt();

                        int res = luaL_dostring( L_ptr_, msg.c_str() );
                        if ( res != 0 ) {
                                SCHPIN_ERROR_LOG( "console", lua_tostring( L_ptr_, -1 ) );
                                lua_pop( L_ptr_, 1 );
                        }

                        ros::spinOnce();
                }
        }
};

int main( int argc, char* argv[] )
{
        ros::init( argc, argv, "console" );
        ros::NodeHandle handle{ "~" };

        ros::ServiceClient def_client =
            handle.serviceClient< schpin_msgs::LuaDef >( "/control/lua/def" );

        schpin_msgs::LuaDef srv;
        if ( !def_client.call( srv ) ) {
                SCHPIN_ERROR_LOG( "console", "Failed to call def service" );
                std::exit( 1 );
        }

        console con{ handle, nlohmann::json::parse( srv.response.definition ) };
        con.run();
}
