#include "schpin_lib/model/leg.h"

#pragma once

namespace schpin
{

template < unsigned N >
class revolute_joint_angle_cache
{
        std::array< em::angle, N > base_;
        std::array< em::angle, N > step_;

      public:
        explicit revolute_joint_angle_cache( const std::array< const joint*, N >& mapping )
        {
                for ( std::size_t i : em::range( N ) ) {
                        mapping[i]->jclass << match(
                            [&]( const revolute_joint& rj ) {
                                    base_[i] = rj.min_angle;
                                    step_[i] = ( rj.max_angle - rj.min_angle ) /
                                               float( rj.state_space_steps - 1 );
                            },
                            [&]( const fixed_joint& ) { SCHPIN_ASSERT( false ); } );
                }
        }

        constexpr std::array< em::angle, N >
        discrete_to_angles( const std::array< uint8_t, N >& angles ) const
        {
                return em::map_f_to_a< N >( em::range( N ), [&]( std::size_t i ) {
                        return base_[i] + step_[i] * static_cast< float >( angles[i] );
                } );
        }
};

}  // namespace schpin
