#include "schpin_lib_exp/overload.h"
#include "schpin_lib_exp/sized_array.h"

#pragma once

namespace schpin
{

template < typename SizeType >
struct revolute_joints_dataset
{

        template < typename T >
        using container = sized_array< T, SizeType >;

        container< const joint* >        joints;   // mapping angle -> joint
        container< pose< robot_frame > > offsets;  // offset of revolute joint from it's first
        container< vec< 3 > >            axis;     // axis for convenience

        revolute_joints_dataset( pose< robot_frame >              first_offset,
                                 const container< const joint* >& mapping )
          : joints( mapping )
          , offsets( joints.get_size_token() )
          , axis( joints.get_size_token() )
        {
                offsets[0] = first_offset;
                for ( std::size_t i : em::range( joints.size() ) ) {
                        find_revolute_children(
                            mapping[i], [&]( const joint*, const pose< robot_frame >& offset ) {
                                    offsets[i + 1] = offset;
                            } );

                        // TODO: this throws exception, deal with that
                        axis[i] = std::get< revolute_joint >( joints[i]->jclass ).axis;
                }
        }

      private:
        template < typename OnChildFunction >
        void find_revolute_children( const joint*        j,
                                     OnChildFunction     f,
                                     pose< robot_frame > parent_offset = {} ) const
        {
                const std::unique_ptr< joint >& child = j->sub_link.child_ptr;
                if ( !child ) {
                        return;
                }
                child->jclass << match(
                    [&]( const revolute_joint& ) {  //
                            f( &*child, parent_offset + child->offset );
                    },
                    [&]( const fixed_joint& ) {  //
                            find_revolute_children( &*child, f, parent_offset + child->offset );
                    } );
        }
};

}  // namespace schpin
