#include <tuple>
#include <variant>

#pragma once

namespace schpin
{

template < class... Fs >
struct overloaded : Fs...
{
        overloaded( const overloaded& )     = default;
        overloaded( overloaded& )           = default;
        overloaded( overloaded&& ) noexcept = default;

        overloaded& operator=( const overloaded& ) = default;
        overloaded& operator=( overloaded& ) = default;
        overloaded& operator=( overloaded&& ) noexcept = default;

        template < typename... Ts >
        explicit overloaded( Ts&&... ts )
          : Fs{ std::forward< Ts >( ts ) }...
        {
        }

        using Fs::operator()...;

        ~overloaded() = default;
};

template < typename... Fs >
auto match( Fs&&... fs )
{
        return overloaded< std::decay_t< Fs >... >{ std::forward< Fs >( fs )... };
}

template < typename... Fs >
auto match_var( Fs&&... fs )
{
        return [m = match( std::forward< Fs >( fs )... )]( auto&& var ) { return var << m; };
}

template < typename... Fs >
auto tuple_match( Fs&&... fs )
{
        return match( [m = match( std::forward< Fs >( fs )... )]( auto&& item ) {
                return std::apply( m, std::forward< decltype( item ) >( item ) );
        } );
}

template < typename... Fs >
auto variant_tuple_match( Fs... fs )
{
        return match( [m = match( std::forward< Fs >( fs )... )]( auto&& item ) {
                return std::visit(
                    [&]( auto&& v ) {  //
                            return std::apply( m, std::forward< decltype( v ) >( v ) );
                    },
                    std::forward< decltype( item ) >( item ) );
        } );
}

template < typename... Ts, typename... Fs >
auto operator<<( const std::variant< Ts... >& v, const overloaded< Fs... >& callable )
{
        return std::visit( callable, v );
}

}  // namespace schpin
