#include <array>
#include <emlabcpp/algorithm.h>
#include <iomanip>
#include <ios>
#include <schpin_lib_exp/convert_iterator.h>

#pragma once

namespace schpin
{

template < typename MsgCounterType, std::size_t N, typename Owner >
class prot_message
{
        std::array< uint8_t, N > data_ = { 0 };
        std::size_t              used_ = 0;

        template < typename Iterator >
        prot_message( Iterator begin, Iterator end )
          : used_( static_cast< std::size_t >( std::distance( begin, end ) ) )
        {
                std::copy( begin, end, data_.begin() );
        }

      public:
        using owner            = Owner;
        using msg_counter_type = MsgCounterType;
        using value_type       = uint8_t;
        using iterator         = uint8_t*;
        using const_iterator   = const uint8_t*;

        static constexpr std::size_t max_size = N;

        template < typename Container >
        static std::optional< prot_message > make( Container&& cont )
        {
                if ( cont.size() > N ) {
                        return {};
                }
                return { prot_message( cont.begin(), cont.end() ) };
        }

        prot_message() = default;

        // TODO: maybe make this explicit?
        template < typename OtherMsgCounterType, std::size_t M, typename OtherOwner >
        explicit prot_message( prot_message< OtherMsgCounterType, M, OtherOwner > other )
          : used_( other.size() )
        {
                static_assert( M <= N );
                std::copy( other.begin(), other.end(), data_.begin() );
        }

        template < std::size_t M >
        explicit prot_message( const std::array< uint8_t, M >& inpt, std::size_t used = M ) noexcept
          : used_( used )
        {
                std::copy( inpt.begin(), inpt.end(), data_.begin() );
                static_assert( M <= N );
        }

        std::size_t size() const noexcept
        {
                return used_;
        }

        uint8_t* begin() noexcept
        {
                return data_.begin();
        }
        uint8_t* end() noexcept
        {
                return data_.begin() + used_;
        }
        const uint8_t* begin() const noexcept
        {
                return data_.begin();
        }
        const uint8_t* end() const noexcept
        {
                return data_.begin() + used_;
        }

        uint8_t& operator[]( std::size_t i ) noexcept
        {
                return data_[i];
        }

        uint8_t operator[]( std::size_t i ) const noexcept
        {
                return data_[i];
        }

        std::string stringify() const
        {
                return Owner::stringify( *this );
        }
};

template < typename MsgCounterType, std::size_t N, typename Owner >
constexpr bool operator==( const prot_message< MsgCounterType, N, Owner >& lh,
                           const prot_message< MsgCounterType, N, Owner >& rh )
{
        return em::equal( lh, rh );
}

template < typename MsgCounterType, std::size_t N, typename Owner >
inline std::ostream& operator<<( std::ostream&                                   os,
                                 const prot_message< MsgCounterType, N, Owner >& m )
{
        return os << convert_view< int >( m );
}

}  // namespace schpin
