#include <schpin_lib_exp/protocol/base.h>

#pragma once

namespace schpin
{

template < auto ID, typename T >
struct protocol_reg
{
        static constexpr auto id = ID;
        using value_type         = T;

        using prot_item                   = protocol_item< value_type >;
        static constexpr std::size_t size = prot_item::size;

        T value;
};

template < typename... >
struct protocol_reg_storage;

template < auto ID >
struct protocol_reg_storage_tag
{
};

template <>
struct protocol_reg_storage<>
{

        template < auto ID >
        static constexpr bool contains_id = false;

        template < auto ID >
        using value_t = void;

        static constexpr std::size_t get_size( std::size_t )
        {
                // maybe assert that this does not happen?
                return 0;
        }

        void get_val() const
        {
        }

        void set_val() const
        {
        }

        template < typename T, typename Buffer >
        opt_error< protocol_error_record > set_raw_val( T, const Buffer& )
        {
                return { protocol_error_record{ UNKNOWN_REG_E, 0 } };
        }
};

template < typename Reg, typename... Regs >
struct protocol_reg_storage< Reg, Regs... > : public protocol_reg_storage< Regs... >
{
        using sub       = protocol_reg_storage< Regs... >;
        using enum_type = decltype( Reg::id );

        Reg reg_;

        template < auto ID >
        static constexpr bool contains_id = ID == Reg::id ? true : sub::template contains_id< ID >;

        template < auto ID >
        using value_t = std::conditional_t< ID == Reg::id,
                                            typename Reg::value_type,
                                            typename sub::template value_t< ID > >;

        static constexpr std::size_t get_size( std::size_t i )
        {
                if ( i == 0 ) {
                        return Reg::size;
                }
                return sub::get_size( i - 1 );
        }

        static constexpr std::size_t get_size( protocol_reg_storage_tag< Reg::id > )
        {
                return Reg::size;
        }
        using sub::get_size;

        typename Reg::value_type get_val( protocol_reg_storage_tag< Reg::id > ) const
        {
                return reg_.value;
        }
        using sub::get_val;

        template < typename Buffer >
        opt_error< protocol_error_record > set_raw_val( enum_type reg, const Buffer& buff )
        {
                opt_error< protocol_error_record > res;
                if ( reg != Reg::id ) {
                        return sub::set_raw_val( reg, buff );
                }
                protocol_item< typename Reg::value_type >::deserialize(
                    buff,
                    match(
                        [&]( value_tag, auto val ) {
                                reg_.value = val;
                        },
                        [&]( error_tag, protocol_error_record rec ) {
                                res = rec;
                        } ) );

                return res;
        }

        void set_val( protocol_reg_storage_tag< Reg::id >, typename Reg::value_type val )
        {
                reg_.value = val;
        }
        using sub::set_val;
};

template < typename Enum, typename... Regs >
class protocol_map
{
        using storage = protocol_reg_storage< Regs... >;

        storage stor_;

        public:
        static constexpr std::size_t                    item_count = sizeof...( Regs );
        static constexpr std::array< Enum, item_count > ids        = { Regs::id... };

        template < auto ID >
        using reg_value_t = typename storage::template value_t< ID >;

        template < auto ID >
        static constexpr bool contains_id = storage::template contains_id< ID >;

        template < auto ID >
        static constexpr std::size_t get_size()
        {
                static_assert( contains_id< ID > );
                return storage::get_size( protocol_reg_storage_tag< ID >{} );
        }

        static constexpr std::size_t get_size( std::size_t i )
        {
                return storage::get_size( i );
        }

        template < auto ID >
        reg_value_t< ID > get_val() const
        {
                static_assert( contains_id< ID > );
                return stor_.get_val( protocol_reg_storage_tag< ID >{} );
        }

        template < typename Buffer >
        opt_error< protocol_error_record > set_val( Enum reg, const Buffer& buff )
        {
                // TODO: proper error handling
                return stor_.set_raw_val( reg, buff );
        }

        template < auto ID >
        void set_val( reg_value_t< ID > val )
        {
                static_assert( contains_id< ID > );
                stor_.set_val( protocol_reg_storage_tag< ID >{}, val );
        }
};

}  // namespace schpin
