#include <ros/ros.h>
#include <schpin_lib/algorithm.h>
#include <visualization_msgs/Marker.h>
#include <visualization_msgs/MarkerArray.h>

#pragma once

namespace schpin
{

class marker_manager
{
        std::vector< std::pair< std::string, int > > robot_active_markers_{};
        ros::Publisher                               marker_array_pub_;

      public:
        marker_manager( ros::NodeHandle handle = { "~" } )
          : marker_array_pub_( handle.advertise< visualization_msgs::MarkerArray >( "/visualization"
                                                                                    "_marker_array",
                                                                                    10 ) )
        {
        }

        void publish( const visualization_msgs::MarkerArray& msg )
        {
                marker_array_pub_.publish( msg );
                for ( const visualization_msgs::Marker& submsg : msg.markers ) {
                        robot_active_markers_.emplace_back( submsg.ns, submsg.id );
                }
        }

        void clear()
        {

                visualization_msgs::MarkerArray arr;
                arr.markers = map_f_to_v(
                    robot_active_markers_, em::uncurry | [&]( std::string ns, int id ) {
                            visualization_msgs::Marker msg;
                            msg.ns     = std::move( ns );
                            msg.id     = id;
                            msg.action = visualization_msgs::Marker::DELETE;
                            return msg;
                    } );
                marker_array_pub_.publish( arr );
        }

        ~marker_manager()
        {
                clear();
        }
};

inline visualization_msgs::Marker create_marker( const std::string& ns,
                                                 unsigned           id,
                                                 const std::string& frame )
{
        visualization_msgs::Marker res;

        res.action          = visualization_msgs::Marker::ADD;
        res.frame_locked    = 1u;
        res.header.frame_id = frame;
        res.header.stamp    = ros::Time();
        res.ns              = ns;
        res.id              = int( id );

        res.pose.position.x    = 0;
        res.pose.position.y    = 0;
        res.pose.position.z    = 0;
        res.pose.orientation.x = 0.0;
        res.pose.orientation.y = 0.0;
        res.pose.orientation.z = 0.0;
        res.pose.orientation.w = 1.0;

        res.color.a = 1.f;

        res.lifetime = ros::Duration();

        return res;
}

}  // namespace schpin
