#include "body/base.h"
#include "graph/generics.h"
#include "graph/generics/edge.h"
#include "graph/generics/vertex.h"
#include "schpin_lib/geom/pose.h"
#include "schpin_lib/geom/pose/serialization.h"

#pragma once

namespace schpin
{

struct body_edge_content
{
        body_price price;
        body_price get_price() const
        {
                return price;
        }
};
using body_edge =
    generic_edge< body_edge_id, body_edge_content, body_price, body_vertex_id, body_edge_offset >;

inline void to_json( nlohmann::json& j, const body_edge_content& content )
{
        j = content.price;
}

inline void from_json( const nlohmann::json& j, body_edge_content& content )
{
        content.price = j.get< body_price >();
}

struct body_vertex_content
{
        pose< world_frame > pos;
};

inline void to_json( nlohmann::json& j, const body_vertex_content& content )
{
        j = content.pos;
}

inline void from_json( const nlohmann::json& j, body_vertex_content& content )
{
        content.pos = j.get< pose< world_frame > >();
}

using body_vertex = generic_vertex< body_vertex_id, body_vertex_content, body_edge >;
using body_graph  = generic_graph< body_vertex, body_edge >;
}  // namespace schpin
