#include "graph/generics.h"
#include "graph/generics/edge.h"
#include "graph/generics/vertex.h"
#include "leg/base.h"
#include "leg/graph/base.h"

#pragma once

namespace schpin
{
struct leg_edge_content
{
        leg_price price;
        leg_price get_price() const
        {
                return price;
        }
};
using leg_edge =
    generic_edge< leg_edge_id, leg_edge_content, leg_price, leg_vertex_id, leg_edge_offset >;
}  // namespace schpin

template <>
struct std::hash< schpin::leg_edge_content >
{
        std::size_t operator()( const schpin::leg_edge_content& content )
        {
                return std::hash< schpin::leg_price >()( content.price );
        }
};

namespace schpin
{
template < unsigned N >
struct leg_vertex_content
{
        discrete_angles< N > angles;
};

template < unsigned N >
using leg_vertex = generic_vertex< leg_vertex_id, leg_vertex_content< N >, leg_edge >;

template < unsigned N >
using leg_graph = generic_graph< leg_vertex< N >, leg_edge >;

}  // namespace schpin

template < unsigned N >
struct std::hash< schpin::leg_vertex_content< N > >
{
        std::size_t operator()( const schpin::leg_vertex_content< N >& content )
        {
                return std::hash< schpin::discrete_angles< N > >()( content.angles );
        }
};

template <>
struct nlohmann::adl_serializer< schpin::leg_edge_content >
{
        static void to_json( nlohmann::json& j, const schpin::leg_edge_content& content )
        {
                j["price"] = content.get_price();
        }

        static schpin::leg_edge_content from_json( const nlohmann::json& j )
        {
                return {j.at( "price" ).get< schpin::leg_price >()};
        }
};

template < unsigned N >
struct nlohmann::adl_serializer< schpin::leg_vertex_content< N > >
{
        static void to_json( nlohmann::json& j, const schpin::leg_vertex_content< N >& cont )
        {
                j["angles"] = cont.angles;
        }

        static schpin::leg_vertex_content< N > from_json( const nlohmann::json& j )
        {
                return {j.at( "angles" ).get< schpin::discrete_angles< N > >()};
        }
};
