#include "octree/proxy_node.h"

#pragma once

namespace schpin
{

// Interface of types that are content of tree
template < typename Content, typename TokenT >
class i_oc_content  // NOLINT(cppcoreguidelines-special-member-functions)
{
      public:
        using token = TokenT;

        virtual std::size_t get_memory_usage( const oc_node< Content >& node ) const = 0;
        virtual bool        prunnable( const oc_node< Content >& node ) const        = 0;
        virtual void        child_based_update( const oc_node< Content >& node )     = 0;
        virtual void        set_prunned( const oc_node< Content >& node )            = 0;
        virtual void        update( const token& token )                             = 0;
        virtual ~i_oc_content()                                                      = default;
};

// this is ugly part of interface... :)
template < typename Content, typename Msg >
oc_node< Content > msg_to_octree_node( const Msg& msg );

template < typename Content, typename MsgT >
class i_oc_serializable_content  // NOLINT(cppcoreguidelines-special-member-functions)
{
      public:
        using msg = MsgT;

        virtual msg to_msg( const oc_node< Content >& node ) const = 0;
        virtual ~i_oc_serializable_content()                       = default;
};

// CollissionCheckers are helper classes designed to implement various detection modes for mechanism
// that modifies tree, based on it's collision with other element. These are used by adequate
// visitors, how?
//
// <name>CollisionVisitor goes through provided item and tree in question, and modifies each node of
// tree, which is in collision with the provided item. (How depends on the content type of the
// tree). In case tree nodes are not present, they are created.
//
// The way the collision is checked depends on the 'CollisionChecker' which can use various ways to
// detect collision
template < typename NodeContentT, typename TagT >
class i_oc_collision_checker  // NOLINT(cppcoreguidelines-special-member-functions)
{
      public:
        using tag          = TagT;
        using node_content = NodeContentT;

        virtual bool
        inner_collission( const oc_node_full_collision_query< node_content, tag >& ) const = 0;
        virtual bool
        bottom_collission( const oc_node_full_collision_query< node_content, tag >& ) const = 0;
        virtual ~i_oc_collision_checker() = default;
};

template < typename InfoType, typename NodeContent, typename Tag >
class i_oc_visitor  // NOLINT(cppcoreguidelines-special-member-functions)
{
      public:
        virtual void access_bottom_leaf( const oc_proxy_node< NodeContent, Tag >&,
                                         const InfoType&,
                                         NodeContent& content ) const = 0;
        virtual std::pair< oc_node_control, InfoType >
        access_node( const oc_proxy_node< NodeContent, Tag >&, const InfoType& ) const = 0;
        virtual std::pair< oc_locked_leaf_control, InfoType >
        access_locked_leaf( const oc_proxy_node< NodeContent, Tag >&, const InfoType& ) const = 0;
        virtual ~i_oc_visitor() = default;
};

template < typename NodeContent, typename Tag >
class i_oc_simple_visitor  // NOLINT(cppcoreguidelines-special-member-functions)
{
      public:
        virtual void            access_bottom_leaf( const oc_proxy_node< NodeContent, Tag >&,
                                                    NodeContent& content ) const              = 0;
        virtual oc_node_control access_node( const oc_proxy_node< NodeContent, Tag >& ) const = 0;
        virtual oc_locked_leaf_control
        access_locked_leaf( const oc_proxy_node< NodeContent, Tag >& ) const = 0;
        virtual ~i_oc_simple_visitor()                                       = default;
};

template < typename NodeContent, typename Tag >
class oc_passive_visitor : public i_oc_simple_visitor< NodeContent, Tag >
{
      public:
        void access_bottom_leaf( const oc_proxy_node< NodeContent, Tag >& /*unused*/,
                                 NodeContent& /*unused*/ ) const final
        {
        }
        oc_node_control
        access_node( const oc_proxy_node< NodeContent, Tag >& /*unused*/ ) const final
        {
                return oc_node_control::FOLLOW_CHILDREN;
        }
        oc_locked_leaf_control
        access_locked_leaf( const oc_proxy_node< NodeContent, Tag >& /*unused*/ ) const final
        {
                return oc_locked_leaf_control::END;
        }
};

struct oc_empty_info_type
{
};

template < typename SimpleVisitor, typename NodeContent, typename Tag >
class oc_simple_visitor_overlay : public i_oc_visitor< oc_empty_info_type, NodeContent, Tag >
{
        const SimpleVisitor& vis_;

      public:
        explicit oc_simple_visitor_overlay( const SimpleVisitor& vis )
          : vis_( vis )
        {
        }
        void access_bottom_leaf( const oc_proxy_node< NodeContent, Tag >& node,
                                 const oc_empty_info_type& /*unused*/,
                                 NodeContent& content ) const final
        {
                vis_.access_bottom_leaf( node, content );
        }
        std::pair< oc_node_control, oc_empty_info_type >
        access_node( const oc_proxy_node< NodeContent, Tag >& node,
                     const oc_empty_info_type& /*unused*/ ) const final
        {
                return {vis_.access_node( node ), oc_empty_info_type{}};
        }
        std::pair< oc_locked_leaf_control, oc_empty_info_type >
        access_locked_leaf( const oc_proxy_node< NodeContent, Tag >& node,
                            const oc_empty_info_type& /*unused*/ ) const final
        {
                return {vis_.access_locked_leaf( node ), oc_empty_info_type{}};
        }
};

}  // namespace schpin
