#include "body/base.h"
#include "environment.h"
#include "robot.h"
#include "schpin_lib_exp/event_reactor.h"

#pragma once

namespace schpin
{

struct body_path_failure_event
{
        std::vector< body_vertex_id > vertexes;
        std::vector< body_edge_id >   edges;
};
struct make_vertexes_event
{
        static constexpr auto     priority = event_priority::DEFAULT;
        unsigned                  count;
        std::optional< unsigned > opt_verbosity{};
};

struct planner_context
{
        pose< world_frame >                  current;
        std::optional< pose< world_frame > > goal;
        environment                          env;
        robot                                rob;

        explicit planner_context( robot robot )
          : rob( std::move( robot ) )
        {
        }

        planner_context( pose< world_frame >                  current,
                         std::optional< pose< world_frame > > goal,
                         environment                          env,
                         robot                                robot )
          : current( current )
          , goal( goal )
          , env( std::move( env ) )
          , rob( std::move( robot ) )
        {
        }
};
}  // namespace schpin

template <>
struct std::hash< schpin::planner_context >
{
        std::size_t operator()( const schpin::planner_context& context )
        {
                return std::hash< schpin::pose< schpin::world_frame > >()( context.current ) ^
                       std::hash< std::optional< schpin::pose< schpin::world_frame > > >()(
                           context.goal ) ^
                       std::hash< schpin::environment >()( context.env ) ^
                       std::hash< schpin::robot >()( context.rob );
        }
};
