#include "leg/sidepath.h"
#include "schpin_test/test_environment.h"

#include <gtest/gtest.h>

using namespace schpin;

extern test_data_archive ARCHIVE;

struct is_surface_predicate_counter
{
        std::size_t checked_counter = 0;
        std::size_t is_surface      = 0;

        constexpr bool operator()( const oc_node< environment_oc_tree_content >& node )
        {
                checked_counter += 1;
                bool res = node->is_surface();
                if ( res ) {
                        is_surface += 1;
                }
                return res;
        }

        template < typename Tag >
        constexpr bool operator()( const oc_proxy_node< environment_oc_tree_content, Tag >& node )
        {
                return this->operator()( *node );
        }
};

TEST( Sidepath, eligible_surface_collection )
{
        const robot&       robot = ARCHIVE.get_robot( "koke" ).copy();
        const environment& env   = ARCHIVE.get_env( "plane" ).copy();
        const leg< 3 >&    leg   = robot.get_legs()[leg_id{ 0 }];

        is_surface_predicate_counter pred;
        auto                         opt_env_box = get_used_space( env.get_tree(), pred );
        auto                         opt_workarea_box =
            get_used_space( leg.get_workarea(), occupancy_treshold_predicate{} );

        EXPECT_TRUE( opt_env_box )
            << "counters: " << pred.checked_counter << "/" << pred.is_surface << "\n"
            << "root: " << env.get_tree().get_proxy_root();
        EXPECT_TRUE( opt_workarea_box );
}

TEST( Sidepath, createGenerator )
{
        const robot&       robot = ARCHIVE.get_robot( "koke" ).copy();
        const environment& env   = ARCHIVE.get_env( "plane" ).copy();
        const leg< 3 >&    leg   = robot.get_legs()[leg_id{ 0 }];

        std::vector< pose< world_frame > > base_path;
        base_path.emplace_back( pose{ point< 3, world_frame >{ -0.1f, 0.f, 0.2f } } );
        base_path.emplace_back( pose{ point< 3, world_frame >{ 0.1f, 0.f, 0.2f } } );

        const std::vector< pose< world_frame > > body_poses =
            lineary_interpolate_path( base_path, em::distance{ 0.05f }, em::angle{ 0.1f } );

        const point< 3, world_frame >& tip_position =
            transform( leg.tip_position( leg.get_model().neutral_configuration() ), body_poses[0] );

        create_sidepath_generator( body_poses, leg, tip_position, env )
            .match(
                [&]( leg_sidepath_generator gen ) {                               //
                        EXPECT_GT( *gen.get_graph().size(), body_poses.size() );  // rough estimate
                        auto p = gen.find_sidepath( leg, env );
                        EXPECT_GT( p.size(), body_poses.size() / 2 );
                        EXPECT_LT( p.size(), body_poses.size() * 2 );
                },
                [&]( const sidepath_failure_report& rep ) { FAIL() << rep; } );
}
